@@ -3,7 +3,7 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2003 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2004 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -29,19 +29,18 @@ extern double pow P((double x, double y));
 extern double modf P((double x, double *yp));
 extern double fmod P((double x, double y));
 
+static inline void make_scalar P((NODE *tree));
 static int eval_condition P((NODE *tree));
 static NODE *op_assign P((NODE *tree));
 static NODE *func_call P((NODE *tree));
 static NODE *match_op P((NODE *tree));
-static int forloops_active P((void));
 static void pop_forloop P((void));
-static void pop_all_forloops P((void));
+static inline void pop_all_forloops P((void));
 static void push_forloop P((const char *varname, NODE **elems, size_t nelems));
 static void push_args P((int count, NODE *arglist, NODE **oldstack,
 			const char *func_name, char **varnames));
-static void pop_fcall_stack P((void));
+static inline void pop_fcall_stack P((void));
 static void pop_fcall P((void));
-static int in_function P((void));
 static int comp_func P((const void *p1, const void *p2));
 
 #if __GNUC__ < 2
@@ -92,9 +91,16 @@ extern int exiting, exit_val;
  *
  * Do NOT make this array static, it is used in several spots, not
  * just in this file.
+ *
+ * 6/2004:
+ * This table is also used for IGNORECASE for == and !=, and index().
+ * Although with GLIBC, we could use tolower() everywhere and RE_ICASE
+ * for the regex matcher, precomputing this table once gives us a
+ * performance improvement.  I also think it's better for portability
+ * to non-GLIBC systems.  All the world is not (yet :-) GNU/Linux.
  */
 #if 'a' == 97	/* it's ascii */
-const char casetable[] = {
+char casetable[] = {
 	'\000', '\001', '\002', '\003', '\004', '\005', '\006', '\007',
 	'\010', '\011', '\012', '\013', '\014', '\015', '\016', '\017',
 	'\020', '\021', '\022', '\023', '\024', '\025', '\026', '\027',
@@ -148,6 +154,33 @@ const char casetable[] = {
 
 #undef C
 
+/* load_casetable --- for a non-ASCII locale, redo the table */
+
+void
+load_casetable(void)
+{
+#if defined(LC_CTYPE)
+	int i;
+	char *cp;
+	static int loaded = FALSE;
+
+	if (loaded || do_traditional)
+		return;
+
+	loaded = TRUE;
+	cp = setlocale(LC_CTYPE, NULL);
+
+	/* this is not per standard, but it's pretty safe */
+	if (cp == NULL || strcmp(cp, "C") == 0 || strcmp(cp, "POSIX") == 0)
+		return;
+
+	for (i = 0200; i <= 0377; i++) {
+		if (isalpha(i) && islower(i) && i != toupper(i))
+			casetable[i] = toupper(i);
+	}
+#endif
+}
+
 /*
  * This table maps node types to strings for debugging.
  * KEEP IN SYNC WITH awk.h!!!!
@@ -176,6 +209,7 @@ static const char *const nodetypes[] = {
 	"Node_assign_plus",
 	"Node_assign_minus",
 	"Node_assign_exp",
+	"Node_assign_concat",
 	"Node_and",
 	"Node_or",
 	"Node_equal",
@@ -190,10 +224,15 @@ static const char *const nodetypes[] = {
 	"Node_rule_list",
 	"Node_rule_node",
 	"Node_statement_list",
+	"Node_switch_body",
+	"Node_case_list",
 	"Node_if_branches",
 	"Node_expression_list",
 	"Node_param_list",
 	"Node_K_if",
+	"Node_K_switch",
+	"Node_K_case",
+	"Node_K_default",
 	"Node_K_while",	
 	"Node_K_for",
 	"Node_K_arrayfor",
@@ -217,6 +256,7 @@ static const char *const nodetypes[] = {
 	"Node_redirect_pipein",
 	"Node_redirect_input",
 	"Node_redirect_twoway",
+	"Node_var_new",
 	"Node_var",
 	"Node_var_array",
 	"Node_val",
@@ -244,6 +284,7 @@ static const char *const nodetypes[] = {
 	"Node_OFS",
 	"Node_ORS",
 	"Node_RS",
+	"Node_SUBSEP",
 	"Node_TEXTDOMAIN",
 	"Node_final --- this should never appear",
 	NULL
@@ -278,11 +319,9 @@ flags2str(int flagval)
 		{ NUMBER, "NUMBER" },
 		{ MAYBE_NUM, "MAYBE_NUM" },
 		{ ARRAYMAXED, "ARRAYMAXED" },
-		{ SCALAR, "SCALAR" },
 		{ FUNC, "FUNC" },
 		{ FIELD, "FIELD" },
 		{ INTLSTR, "INTLSTR" },
-		{ UNINITIALIZED, "UNINITIALIZED" },
 		{ 0,	NULL },
 	};
 
@@ -301,15 +340,15 @@ genflags2str(int flagval, const struct flagtab *tab)
 	sp = buffer;
 	space_left = BUFSIZ;
 	for (i = 0; tab[i].name != NULL; i++) {
-		/*
-		 * note the trick, we want 1 or 0 for whether we need
-		 * the '|' character.
-		 */
-		space_needed = (strlen(tab[i].name) + (sp != buffer));
-		if (space_left < space_needed)
-			fatal(_("buffer overflow in genflags2str"));
-
 		if ((flagval & tab[i].val) != 0) {
+			/*
+			 * note the trick, we want 1 or 0 for whether we need
+			 * the '|' character.
+			 */
+			space_needed = (strlen(tab[i].name) + (sp != buffer));
+			if (space_left < space_needed)
+				fatal(_("buffer overflow in genflags2str"));
+
 			if (sp != buffer) {
 				*sp++ = '|';
 				space_left--;
@@ -324,6 +363,49 @@ genflags2str(int flagval, const struct flagtab *tab)
 	return buffer;
 }
 
+/*
+ * make_scalar --- make sure that tree is a scalar.
+ *
+ * tree is in a scalar context.  If it is a variable, accomplish
+ * what's needed; otherwise, do nothing.
+ *
+ * Notice that nodes of type Node_var_new have undefined value in var_value
+ * (a.k.a. lnode)---even though awkgram.y:variable() initializes it,
+ * push_args() doesn't.  Thus we have to initialize it.
+ */
+
+static inline void
+make_scalar P((NODE *tree))
+{
+	switch (tree->type) {
+	case Node_var_array:
+		fatal(_("attempt to use array `%s' in a scalar context"),
+			array_vname(tree));
+
+	case Node_array_ref:
+		switch (tree->orig_array->type) {
+		case Node_var_array:
+			fatal(_("attempt to use array `%s' in a scalar context"),
+				array_vname(tree));
+		case Node_var_new:
+			tree->orig_array->type = Node_var;
+			tree->orig_array->var_value = Nnull_string;
+			break;
+		case Node_var:
+			break;
+		default:
+			cant_happen();
+		}
+		/* fall through */
+	case Node_var_new:
+		tree->type = Node_var;
+		tree->var_value = Nnull_string;
+	default:
+		/* shut up GCC */
+		break;
+	}
+}
+
 /*
  * interpret:
  * Tree is a bunch of rules to run. Returns zero if it hit an exit()
@@ -372,16 +454,12 @@ interpret(register NODE *volatile tree)
 				}
 				break;
 			case TAG_CONTINUE:	/* NEXT statement */
-				if (forloops_active())
-					pop_all_forloops();
-				if (in_function())
-					pop_fcall_stack();
+				pop_all_forloops();
+				pop_fcall_stack();
 				return 1;
 			case TAG_BREAK:		/* EXIT statement */
-				if (forloops_active())
-					pop_all_forloops();
-				if (in_function())
-					pop_fcall_stack();
+				pop_all_forloops();
+				pop_fcall_stack();
 				return 0;
 			default:
 				cant_happen();
@@ -406,6 +484,109 @@ interpret(register NODE *volatile tree)
 		}
 		break;
 
+	case Node_K_switch:
+		{
+		NODE *switch_value;
+		NODE *switch_body;
+		NODE *case_list;
+		NODE *default_list;
+		NODE *case_stmt;
+
+		int match_found = FALSE;
+
+		PUSH_BINDING(loop_tag_stack, loop_tag, loop_tag_valid);
+		INCREMENT(tree->exec_count);
+		stable_tree = tree;
+
+		switch_value = tree_eval(stable_tree->lnode);
+		switch_body = stable_tree->rnode;
+		case_list = switch_body->lnode;
+		default_list  = switch_body->rnode;
+
+		for (; case_list != NULL; case_list = case_list->rnode) {
+			case_stmt = case_list->lnode;
+
+			/*
+			 * Once a match is found, all cases will be processed as they fall through,
+			 * so continue to execute statements until a break is reached.
+			 */
+			if (! match_found) {
+				if (case_stmt->type == Node_K_default)
+					;	/* do nothing */
+				else if (case_stmt->lnode->type == Node_regex) {
+					NODE *t1;
+					Regexp *rp;
+					/* see comments in match_op() code about this. */
+					int kludge_need_start = FALSE;
+
+					t1 = force_string(switch_value);
+					rp = re_update(case_stmt->lnode);
+
+					if (avoid_dfa(tree, t1->stptr, t1->stlen))
+						kludge_need_start = TRUE;
+					match_found = (research(rp, t1->stptr, 0, t1->stlen, kludge_need_start) >= 0);
+					if (t1 != switch_value)
+						free_temp(t1);
+				} else
+					match_found = (cmp_nodes(switch_value, case_stmt->lnode) == 0);
+			}
+
+			/* If a match was found, execute the statements associated with the case. */
+			if (match_found) {
+				INCREMENT(case_stmt->exec_count);
+				switch (setjmp(loop_tag)) {
+				case 0:                /* Normal non-jump    */
+					(void) interpret(case_stmt->rnode);
+					break;
+				case TAG_CONTINUE:     /* continue statement */
+					free_temp(switch_value);
+					RESTORE_BINDING(loop_tag_stack, loop_tag, loop_tag_valid);
+					longjmp(loop_tag, TAG_CONTINUE);
+					break;
+				case TAG_BREAK:        /* break statement    */
+					free_temp(switch_value);
+					RESTORE_BINDING(loop_tag_stack, loop_tag, loop_tag_valid);
+					return 1;
+				default:
+					cant_happen();
+				}
+			}
+
+		}
+
+		free_temp(switch_value);
+
+		/*
+		 * If a default section was found, execute the statements associated with it
+		 * and execute any trailing case statements if the default falls through.
+		 */
+		if (! match_found && default_list != NULL) {
+			for (case_list = default_list;
+					case_list != NULL; case_list = case_list->rnode) {
+				case_stmt = case_list->lnode;
+
+				INCREMENT(case_stmt->exec_count);
+				switch (setjmp(loop_tag)) {
+				case 0:                /* Normal non-jump    */
+					(void) interpret(case_stmt->rnode);
+					break;
+				case TAG_CONTINUE:     /* continue statement */
+					RESTORE_BINDING(loop_tag_stack, loop_tag, loop_tag_valid);
+					longjmp(loop_tag, TAG_CONTINUE);
+					break;
+				case TAG_BREAK:        /* break statement    */
+					RESTORE_BINDING(loop_tag_stack, loop_tag, loop_tag_valid);
+					return 1;
+				default:
+					cant_happen();
+				}
+			}
+		}
+
+		RESTORE_BINDING(loop_tag_stack, loop_tag, loop_tag_valid);
+		}
+		break;
+
 	case Node_K_while:
 		PUSH_BINDING(loop_tag_stack, loop_tag, loop_tag_valid);
 
@@ -475,8 +656,9 @@ interpret(register NODE *volatile tree)
 	case Node_K_arrayfor:
 		{
 		Func_ptr after_assign = NULL;
-		NODE **list = 0;
+		NODE **list = NULL;
 		NODE *volatile array;
+		NODE *volatile save_array;
 		volatile size_t i, num_elems;
 		size_t j;
 		volatile int retval = 0;
@@ -485,19 +667,12 @@ interpret(register NODE *volatile tree)
 #define hakvar forloop->init
 #define arrvar forloop->incr
 		/* get the array */
-		array = tree->arrvar;
-		if (array->type == Node_param_list)
-			array = stack_ptr[array->param_cnt];
-		if (array->type == Node_array_ref)
-			array = array->orig_array;
-		if ((array->flags & SCALAR) != 0)
-			fatal(_("attempt to use scalar `%s' as array"), array->vname);
+		save_array = tree->arrvar;
+		array = get_array(save_array);
 
 		/* sanity: do nothing if empty */
-		if (array->type == Node_var || array->var_array == NULL
-		    || array->table_size == 0) {
+		if (array->var_array == NULL || array->table_size == 0)
 			break;	/* from switch */
-		}
 
 		/* allocate space for array */
 		num_elems = array->table_size;
@@ -553,7 +728,7 @@ interpret(register NODE *volatile tree)
 
 		if (do_lint && num_elems != array->table_size)
 			lintwarn(_("for loop: array `%s' changed size from %ld to %ld during loop execution"),
-					array->vname, (long) num_elems, (long) array->table_size);
+				array_vname(save_array), (long) num_elems, (long) array->table_size);
 		
 		if (retval == 1)
 			return 1;
@@ -652,11 +827,8 @@ interpret(register NODE *volatile tree)
 		 * Have to do this cleanup here, since we don't longjump
 		 * back to the main awk rule loop (rule_tag).
 		 */
-		if (forloops_active())
-			pop_all_forloops();
-
-		if (in_function())
-			pop_fcall_stack();
+		pop_all_forloops();
+		pop_fcall_stack();
 
 		do_nextfile();
 		break;
@@ -682,8 +854,12 @@ interpret(register NODE *volatile tree)
 	case Node_K_return:
 		INCREMENT(tree->exec_count);
 		t = tree_eval(tree->lnode);
-		ret_node = dupnode(t);
-		free_temp(t);
+		if ((t->flags & (PERM|TEMP)) != 0)
+			ret_node = t;
+		else {
+			ret_node = copynode(t);  /* don't do a dupnode here */
+			ret_node->flags |= TEMP;
+		}
 		longjmp(func_tag, TAG_RETURN);
 		break;
 
@@ -692,7 +868,7 @@ interpret(register NODE *volatile tree)
 		 * Appears to be an expression statement.  Throw away the
 		 * value. 
 		 */
-		if (do_lint && tree->type == Node_var)
+		if (do_lint && (tree->type == Node_var || tree->type == Node_var_new))
 			lintwarn(_("statement has no effect"));
 		INCREMENT(tree->exec_count);
 		t = tree_eval(tree);
@@ -717,20 +893,20 @@ r_tree_eval(register NODE *tree, int iscond)
 	long lx2;
 #endif
 
-#ifdef GAWKDEBUG
+#ifndef TREE_EVAL_MACRO
 	if (tree == NULL)
-		return Nnull_string;
-	else if (tree->type == Node_val) {
+		cant_happen();
+	if (tree->type == Node_val) {
 		if (tree->stref <= 0)
 			cant_happen();
-		return tree;
+		return ((tree->flags & INTLSTR) != 0
+			? r_force_string(tree)
+			: tree);
 	} else if (tree->type == Node_var) {
 		if (tree->var_value->stref <= 0)
 			cant_happen();
-		if ((tree->flags & UNINITIALIZED) != 0)
-			warning(_("reference to uninitialized variable `%s'"), 
-			      tree->vname);
-		return tree->var_value;
+		if (! var_uninitialized(tree))
+			return tree->var_value;
 	}
 #endif
 
@@ -748,16 +924,16 @@ r_tree_eval(register NODE *tree, int iscond)
 			return Nnull_string;
 		}
 
-		if (do_lint && (tree->flags & UNINITIALIZED) != 0)
+		if (do_lint && var_uninitialized(tree))
 			lintwarn(_("reference to uninitialized argument `%s'"),
 			      tree->vname);
 	}
-	if (tree->type == Node_array_ref)
-		tree = tree->orig_array;
+
+	make_scalar(tree);
 
 	switch (tree->type) {
 	case Node_var:
-		if (do_lint && (tree->flags & UNINITIALIZED) != 0)
+		if (do_lint && var_uninitialized(tree))
 			lintwarn(_("reference to uninitialized variable `%s'"),
 			      tree->vname);
 		return tree->var_value;
@@ -778,7 +954,7 @@ r_tree_eval(register NODE *tree, int iscond)
 		return (*tree->builtin)(tree->subnode);
 
 	case Node_K_getline:
-		return (do_getline(tree));
+		return do_getline(tree);
 
 	case Node_in_array:
 		return tmp_number((AWKNUM) (in_array(tree->lnode, tree->rnode) != NULL));
@@ -802,14 +978,11 @@ r_tree_eval(register NODE *tree, int iscond)
 	case Node_CONVFMT:
 	case Node_BINMODE:
 	case Node_LINT:
+	case Node_SUBSEP:
 	case Node_TEXTDOMAIN:
 		lhs = get_lhs(tree, (Func_ptr *) NULL, TRUE);
 		return *lhs;
 
-	case Node_var_array:
-		fatal(_("attempt to use array `%s' in a scalar context"),
-			tree->vname);
-
 	case Node_unary_minus:
 		t1 = tree_eval(tree->subnode);
 		x = -force_number(t1);
@@ -927,6 +1100,60 @@ r_tree_eval(register NODE *tree, int iscond)
 		return r;
 
 	/* assignments */
+	case Node_assign_concat:
+	{
+		Func_ptr after_assign = NULL;
+		NODE *l, *r;
+
+		/*
+		 * Note that something lovely like this:
+		 *
+		 * BEGIN { a = "a"; a = a (a = "b"); print a }
+		 *
+		 * is not defined.  It could print `ab' or `bb'.
+		 * Gawk 3.1.3 prints `ab', so we do that too, simply
+		 * by evaluating the LHS first.  Ugh.
+		 *
+		 * Thanks to mary1john@earthlink.net for pointing
+		 * out this issue.
+		 */
+		lhs = get_lhs(tree->lnode, &after_assign, FALSE);
+		*lhs = force_string(*lhs);
+		l = *lhs;
+		r = force_string(tree_eval(tree->rnode));
+
+		/*
+		 * Don't clobber string constants!
+		 *
+		 * Also check stref; see test/strcat1.awk,
+		 * the test for l->stref == 1 can't be an
+		 * assertion.
+		 *
+		 * Thanks again to mary1john@earthlink.net for pointing
+		 * out this issue.
+		 */
+		if (l != r && (l->flags & PERM) == 0 && l->stref == 1) {
+			size_t nlen = l->stlen + r->stlen + 2;
+
+			erealloc(l->stptr, char *, nlen, "interpret");
+			memcpy(l->stptr + l->stlen, r->stptr, r->stlen);
+			l->stlen += r->stlen;
+		} else {
+			char *nval;
+			size_t nlen = l->stlen + r->stlen + 2;
+
+			emalloc(nval, char *, nlen, "interpret");
+			memcpy(nval, l->stptr, l->stlen);
+			memcpy(nval + l->stlen, r->stptr, r->stlen);
+			unref(*lhs);
+			*lhs = make_str_node(nval, l->stlen + r->stlen, ALREADY_MALLOCED);
+		}
+		free_temp(r);
+
+		if (after_assign)
+			(*after_assign)();
+		return *lhs;
+	}
 	case Node_assign:
 		{
 		Func_ptr after_assign = NULL;
@@ -937,7 +1164,6 @@ r_tree_eval(register NODE *tree, int iscond)
 		lhs = get_lhs(tree->lnode, &after_assign, FALSE);
 
 		assign_val(lhs, r);
-		tree->lnode->flags |= SCALAR;
 		if (after_assign)
 			(*after_assign)();
 		return *lhs;
@@ -1055,7 +1281,7 @@ r_tree_eval(register NODE *tree, int iscond)
 	default:
 		fatal(_("illegal type (%s) in tree_eval"), nodetype2str(tree->type));
 	}
-	return 0;
+	return (NODE *) 0;
 }
 
 /* eval_condition --- is TREE true or false? Returns 0==false, non-zero==true */
@@ -1155,6 +1381,7 @@ cmp_nodes(register NODE *t1, register NODE *t2)
 					     (const char *) cp2, mbs, l);
 		} else
 #endif
+		/* Could use tolower() here; see discussion above. */
 		for (ret = 0; l-- > 0 && ret == 0; cp1++, cp2++)
 			ret = casetable[*cp1] - casetable[*cp2];
 	} else
@@ -1261,7 +1488,6 @@ op_assign(register NODE *tree)
 		cant_happen();
 	}
 
-	tree->lnode->flags |= SCALAR;
 	if (after_assign)
 		(*after_assign)();
 
@@ -1298,15 +1524,6 @@ static struct loop_info {
 size_t nloops = 0;		/* how many slots there are in the stack */
 size_t nloops_active = 0;	/* how many loops are actively stacked */
 
-
-/* forloops_active --- return true if there are loops that need popping */
-
-static int
-forloops_active()
-{
-	return nloops > 0;
-}
-
 /* pop_forloop --- pop one for loop off the stack */
 
 static void
@@ -1332,7 +1549,7 @@ pop_forloop()
 
 /* pop_forloops --- pop the for loops stack all the way */
 
-static void
+static inline void
 pop_all_forloops()
 {
 	while (nloops_active > 0)
@@ -1363,23 +1580,48 @@ push_forloop(const char *varname, NODE **elems, size_t nelems)
 	nloops_active++;
 }
 
+/*
+ * 2/2004:
+ * N.B. The code that uses fcalls[] *always* uses indexing.
+ * This avoids severe problems in case fcalls gets realloc()'ed
+ * during recursive tree_eval()'s or whatever, so that we don't
+ * have to carefully reassign pointers into the array.  The
+ * minor speed gain from using a pointer was offset too much
+ * by the hassles to get the code right and commented.
+ *
+ * Thanks and a tip of the hatlo to Brian Kernighan.
+ */
+
 static struct fcall {
 	const char *fname;	/* function name */
-	unsigned long count;	/* how many args */
+	size_t count;		/* how many args */
 	NODE *arglist;		/* list thereof */
 	NODE **prevstack;	/* function stack frame of previous function */
 	NODE **stack;		/* function stack frame of current function */
-} *fcall_list = NULL;
+} *fcalls = NULL;
 
 static long fcall_list_size = 0;
 static long curfcall = -1;
 
-/* in_function --- return true/false if we need to unwind awk functions */
-
-static int
-in_function()
+/*
+ * get_curfunc_arg_count --- return number actual parameters
+ *
+ * This is for use by dynamically loaded C extension functions.
+ */
+size_t
+get_curfunc_arg_count(void)
 {
-	return (curfcall >= 0);
+	NODE *argp;
+	size_t argc;
+
+	assert(curfcall >= 0);
+
+	/* count the # of expressions in argument expression list */
+	for (argc = 0, argp = fcalls[curfcall].arglist;
+	     argp != NULL; argp = argp->rnode)
+		argc++;
+
+	return argc;
 }
 
 /* pop_fcall --- pop off a single function call */
@@ -1387,79 +1629,32 @@ in_function()
 static void
 pop_fcall()
 {
-	NODE *n, **sp, *arg, *argp;
+	NODE *n, **sp;
 	int count;
-	struct fcall *f;
 
 	assert(curfcall >= 0);
-	f = & fcall_list[curfcall];
-	stack_ptr = f->prevstack;
+	stack_ptr = fcalls[curfcall].prevstack;
 
-	/*
-	 * here, we pop each parameter and check whether
-	 * it was an array.  If so, and if the arg. passed in was
-	 * a simple variable, then the value should be copied back.
-	 * This achieves "call-by-reference" for arrays.
-	 */
-	sp = f->stack;
-	count = f->count;
+	sp = fcalls[curfcall].stack;
 
-	for (argp = f->arglist; count > 0 && argp != NULL; argp = argp->rnode) {
-		arg = argp->lnode;
-		if (arg->type == Node_param_list)
-			arg = stack_ptr[arg->param_cnt];
+	for (count = fcalls[curfcall].count; count > 0; count--) {
 		n = *sp++;
-		if (n->type == Node_var_array || n->type == Node_array_ref) {
-			NODETYPE old_type;	/* for check, below */
-
-			old_type = arg->type;
-
-			/*
-			 * We only free n->vname in the same cases in push_fcall() which
-			 * mallocs it.  These are when passing a real array in; the
-			 * old type is array and the parameter "copy" is an array ref,
-			 * or when passing an array reference on to another function.
-			 * It helps to be awake when writing code.
-			 */
-			if (  (old_type == Node_var_array && n->type == Node_array_ref)
-			   || (old_type == Node_array_ref && n->type == Node_array_ref) )
-				free(n->vname);
-
-			if (arg->type == Node_var) {
-				/* type changed, copy array back for call by reference */
-				/* should we free arg->var_value ? */
-				arg->var_array = n->var_array;
-				arg->type = Node_var_array;
-				arg->array_size = n->array_size;
-				arg->table_size = n->table_size;
-				arg->flags = n->flags;
-			}
-		}
-		/* n->lnode overlays the array size, don't unref it if array */
-		if (n->type != Node_var_array && n->type != Node_array_ref)
-			unref(n->lnode);
-		freenode(n);
-		count--;
-	}
-	while (count-- > 0) {
-		n = *sp++;
-		/* if n is a local array, all the elements should be freed */
-		if (n->type == Node_var_array)
+		if (n->type == Node_var)		/* local variable */
+			unref(n->var_value);
+		else if (n->type == Node_var_array)	/* local array */
 			assoc_clear(n);
-		/* n->lnode overlays the array size, don't unref it if array */
-		if (n->type != Node_var_array && n->type != Node_array_ref)
-			unref(n->lnode);
 		freenode(n);
 	}
-	if (f->stack)
-		free((char *) f->stack);
-	memset(f, '\0', sizeof(struct fcall));
+	if (fcalls[curfcall].stack) {
+		free((char *) fcalls[curfcall].stack);
+		fcalls[curfcall].stack = NULL;
+	}
 	curfcall--;
 }
 
 /* pop_fcall_stack --- pop off all function args, don't leak memory */
 
-static void
+static inline void
 pop_fcall_stack()
 {
 	while (curfcall >= 0)
@@ -1470,99 +1665,73 @@ pop_fcall_stack()
 
 static void
 push_args(int count,
-	NODE *arglist,
+	NODE *argp,
 	NODE **oldstack,
 	const char *func_name,
 	char **varnames)
 {
-	struct fcall *f;
-	NODE *arg, *argp, *r, **sp, *n;
+	NODE *arg, *r, **sp;
 	int i;
-	int num_args;
-
-	num_args = count;	/* save for later use */
 
 	if (fcall_list_size == 0) {	/* first time */
-		emalloc(fcall_list, struct fcall *, 10 * sizeof(struct fcall),
+		emalloc(fcalls, struct fcall *, 10 * sizeof(struct fcall),
 			"push_args");
 		fcall_list_size = 10;
 	}
 
 	if (++curfcall >= fcall_list_size) {
 		fcall_list_size *= 2;
-		erealloc(fcall_list, struct fcall *,
+		erealloc(fcalls, struct fcall *,
 			fcall_list_size * sizeof(struct fcall), "push_args");
 	}
-	f = & fcall_list[curfcall];
-	memset(f, '\0', sizeof(struct fcall));
 
 	if (count > 0)
-		emalloc(f->stack, NODE **, count*sizeof(NODE *), "push_args");
-	f->count = count;
-	f->fname = func_name;	/* not used, for debugging, just in case */
-	f->arglist = arglist;
-	f->prevstack = oldstack;
+		emalloc(fcalls[curfcall].stack, NODE **, count*sizeof(NODE *), "push_args");
+	else
+		fcalls[curfcall].stack = NULL;
+	fcalls[curfcall].count = count;
+	fcalls[curfcall].fname = func_name;	/* not used, for debugging, just in case */
+	fcalls[curfcall].arglist = argp;
+	fcalls[curfcall].prevstack = oldstack;
 
-	sp = f->stack;
+	sp = fcalls[curfcall].stack;
 
 	/* for each calling arg. add NODE * on stack */
-	for (argp = arglist, i = 0; count > 0 && argp != NULL; argp = argp->rnode) {
-		static const char from[] = N_("%s (from %s)");
-		arg = argp->lnode;
+	for (i = 0; i < count; i++) {
 		getnode(r);
-		r->type = Node_var;
-		r->flags = 0;
-		/* call by reference for arrays; see below also */
-		if (arg->type == Node_param_list) {
-			/* we must also reassign f here; see below */
-			f = & fcall_list[curfcall];
-			arg = f->prevstack[arg->param_cnt];
+		*sp++ = r;
+		if (argp == NULL) {
+			/* local variable */
+			r->type = Node_var_new;
+			r->var_value = Nnull_string;
+			r->vname = varnames[i];
+			r->rnode = NULL;
+			continue;
 		}
-		if (arg->type == Node_var_array) {
-			char *p;
-			size_t len;
+		arg = argp->lnode;
+		/* call by reference for arrays; see below also */
+		if (arg->type == Node_param_list)
+			arg = fcalls[curfcall].prevstack[arg->param_cnt];
 
+		if (arg->type == Node_var_array || arg->type == Node_var_new) {
 			r->type = Node_array_ref;
-			r->flags &= ~SCALAR;
 			r->orig_array = arg;
-			len = strlen(varnames[i]) + strlen(arg->vname)
-				+ strlen(gettext(from)) - 4 + 1;
-			emalloc(p, char *, len, "push_args");
-			sprintf(p, _(from), varnames[i], arg->vname);
-			r->vname = p;
+			r->prev_array = arg;
 		} else if (arg->type == Node_array_ref) {
-			char *p;
-			size_t len;
-
   			*r = *arg;
-			len = strlen(varnames[i]) + strlen(arg->vname)
-				+ strlen(gettext(from)) - 4 + 1;
-			emalloc(p, char *, len, "push_args");
-			sprintf(p, _(from), varnames[i], arg->vname);
-			r->vname = p;
+			r->prev_array = arg;
 		} else {
-			n = tree_eval(arg);
+			NODE *n = tree_eval(arg);
+
+			r->type = Node_var;
 			r->lnode = dupnode(n);
 			r->rnode = (NODE *) NULL;
-  			if ((n->flags & SCALAR) != 0)
-	  			r->flags |= SCALAR;
-			r->vname = varnames[i];
 			free_temp(n);
   		}
-		*sp++ = r;
-		i++;
-		count--;
+		r->vname = varnames[i];
+		argp = argp->rnode;
 	}
 
-	/*
-	 * We have to reassign f.  Why, you may ask?  It is possible that
-	 * other functions were called during the course of tree_eval()-ing
-	 * the arguments to this function.  As a result of that, fcall_list
-	 * may have been realloc()'ed, with the result that f is now
-	 * pointing into free()'d space.  This was a nasty one to track down.
-	 */
-	f = & fcall_list[curfcall];
-
 	if (argp != NULL) {
 		/* Left over calling args. */
 		warning(
@@ -1572,30 +1741,15 @@ push_args(int count,
 		do {
 			arg = argp->lnode;
 			if (arg->type == Node_param_list)
-				arg = f->prevstack[arg->param_cnt];
+				arg = fcalls[curfcall].prevstack[arg->param_cnt];
 			if (arg->type != Node_var_array &&
-			    arg->type != Node_array_ref)
+			    arg->type != Node_array_ref &&
+			    arg->type != Node_var_new)
 				free_temp(tree_eval(arg));
-
-			/* reassign f, tree_eval could have moved it */
-			f = & fcall_list[curfcall];
 		} while ((argp = argp->rnode) != NULL);
 	}
 
-	/* add remaining params. on stack with null value */
-	while (count-- > 0) {
-		getnode(r);
-		r->type = Node_var;
-		r->lnode = Nnull_string;
-		r->flags &= ~SCALAR;
-		r->rnode = (NODE *) NULL;
-		r->vname = varnames[i++];
-		r->flags = UNINITIALIZED;
-		r->param_cnt = num_args - count;
-		*sp++ = r;
-	}
-
-	stack_ptr = f->stack;
+	stack_ptr = fcalls[curfcall].stack;
 }
 
 /* func_call --- call a function, call by reference for arrays */
@@ -1674,8 +1828,6 @@ func_call(NODE *tree)
 		RESTORE_BINDING(loop_tag_stack, loop_tag, junk);
 	}
 
-	if ((r->flags & PERM) == 0)
-		r->flags |= TEMP;
 	return r;
 }
 
@@ -1692,7 +1844,7 @@ dump_fcall_stack(FILE *fp)
 
 	fprintf(fp, _("\n\t# Function Call Stack:\n\n"));
 	for (i = curfcall; i >= 0; i--)
-		fprintf(fp, "\t# %3d. %s\n", i+1, fcall_list[i].fname);
+		fprintf(fp, "\t# %3d. %s\n", i+1, fcalls[i].fname);
 	fprintf(fp, _("\t# -- main --\n"));
 }
 #endif /* PROFILING */
@@ -1722,19 +1874,14 @@ r_get_lhs(register NODE *ptr, Func_ptr *assign, int reference)
 		ptr = stack_ptr[ptr->param_cnt];
 	}
 
-	switch (ptr->type) {
-	case Node_array_ref:
-	case Node_var_array:
-		fatal(_("attempt to use array `%s' in a scalar context"),
-			ptr->vname);
+	make_scalar(ptr);
 
+	switch (ptr->type) {
 	case Node_var:
-		if (do_lint && reference && (ptr->flags & UNINITIALIZED) != 0)
+		if (do_lint && reference && var_uninitialized(ptr))
 			lintwarn(_("reference to uninitialized variable `%s'"),
 					      ptr->vname);
 
-		/* clear the flag, since it's about to be assigned to */
-		ptr->flags &= ~UNINITIALIZED;
 		aptr = &(ptr->var_value);
 #ifdef GAWKDEBUG
 		if (ptr->var_value->stref <= 0)
@@ -1783,7 +1930,7 @@ r_get_lhs(register NODE *ptr, Func_ptr *assign, int reference)
 	case Node_NF:
 		if (NF == -1 || NF_node->var_value->numbr != NF) {
 			if (NF == -1)
-				(void) get_field(HUGE-1, assign); /* parse record */
+				(void) get_field(UNLIMITED-1, assign); /* parse record */
 			unref(NF_node->var_value);
 			NF_node->var_value = make_number((AWKNUM) NF);
 		}
@@ -1834,6 +1981,12 @@ r_get_lhs(register NODE *ptr, Func_ptr *assign, int reference)
 			*assign = set_OFS;
 		break;
 
+	case Node_SUBSEP:
+		aptr = &(SUBSEP_node->var_value);
+		if (assign != NULL)
+			*assign = set_SUBSEP;
+		break;
+
 	case Node_TEXTDOMAIN:
 		aptr = &(TEXTDOMAIN_node->var_value);
 		if (assign != NULL)
@@ -1860,23 +2013,16 @@ r_get_lhs(register NODE *ptr, Func_ptr *assign, int reference)
 			aptr = &fields_arr[0];
 			if (assign != NULL)
 				*assign = reset_record;
-			break;
-		}
-		aptr = get_field(field_num, assign);
+		} else
+			aptr = get_field(field_num, assign);
+		if (do_lint && reference && (*aptr == Null_field || *aptr == Nnull_string))
+			lintwarn(_("reference to uninitialized field `$%d'"),
+					      field_num);
 		break;
 		}
 
 	case Node_subscript:
-		n = ptr->lnode;
-		if (n->type == Node_param_list) {
-			n = stack_ptr[n->param_cnt];
-			if ((n->flags & SCALAR) != 0)
-				fatal(_("attempt to use scalar parameter `%s' as an array"), n->vname);
-		}
-		if (n->type == Node_array_ref) {
-			n = n->orig_array;
-			assert(n->type == Node_var_array || n->type == Node_var);
-		}
+		n = get_array(ptr->lnode);
 		aptr = assoc_lookup(n, concat_exp(ptr->rnode), reference);
 		break;
 
@@ -1913,6 +2059,7 @@ match_op(register NODE *tree)
 	register Regexp *rp;
 	int i;
 	int match = TRUE;
+	int kludge_need_start = FALSE;	/* FIXME: --- see below */
 
 	if (tree->type == Node_nomatch)
 		match = FALSE;
@@ -1923,7 +2070,22 @@ match_op(register NODE *tree)
 		tree = tree->rnode;
 	}
 	rp = re_update(tree);
-	i = research(rp, t1->stptr, 0, t1->stlen, FALSE);
+	/*
+	 * FIXME:
+	 *
+	 * Any place where research() is called with a last parameter of
+	 * FALSE, we need to use the avoid_dfa test. This appears here and
+	 * in the code for Node_K_switch.
+	 *
+	 * A new or improved dfa that distinguishes beginning/end of
+	 * string from beginning/end of line will allow us to get rid of
+	 * this temporary hack.
+	 *
+	 * The avoid_dfa() function is in re.c; it is not very smart.
+	 */
+	if (avoid_dfa(tree, t1->stptr, t1->stlen))
+		kludge_need_start = TRUE;
+	i = research(rp, t1->stptr, 0, t1->stlen, kludge_need_start);
 	i = (i == -1) ^ (match == TRUE);
 	free_temp(t1);
 	return tmp_number((AWKNUM) i);
@@ -1940,6 +2102,7 @@ set_IGNORECASE()
 		warned = TRUE;
 		lintwarn(_("`IGNORECASE' is a gawk extension"));
 	}
+	load_casetable();
 	if (do_traditional)
 		IGNORECASE = FALSE;
 	else if ((IGNORECASE_node->var_value->flags & (STRING|STRCUR)) != 0) {
@@ -2039,10 +2202,20 @@ fmt_ok(NODE *n)
 {
 	NODE *tmp = force_string(n);
 	const char *p = tmp->stptr;
+#if ! defined(PRINTF_HAS_F_FORMAT) || PRINTF_HAS_F_FORMAT != 1
+	static const char float_formats[] = "efgEG";
+#else
+	static const char float_formats[] = "efgEFG";
+#endif
+#if ENABLE_NLS && defined(HAVE_LOCALE_H)
+	static const char flags[] = " +-#'";
+#else
+	static const char flags[] = " +-#";
+#endif
 
 	if (*p++ != '%')
 		return 0;
-	while (*p && strchr(" +-#", *p) != NULL)	/* flags */
+	while (*p && strchr(flags, *p) != NULL)	/* flags */
 		p++;
 	while (*p && ISDIGIT(*p))	/* width - %*.*g is NOT allowed */
 		p++;
@@ -2052,7 +2225,7 @@ fmt_ok(NODE *n)
 		p++;
 	while (*p && ISDIGIT(*p))	/* precision */
 		p++;
-	if (*p == '\0' || strchr("efgEG", *p) == NULL)
+	if (*p == '\0' || strchr(float_formats, *p) == NULL)
 		return 0;
 	if (*++p != '\0')
 		return 0;
@@ -2217,6 +2390,7 @@ comp_func(const void *p1, const void *p2)
 	size_t len1, len2;
 	const char *str1, *str2;
 	const NODE *t1, *t2;
+	int cmp1;
 
 	t1 = *((const NODE *const *) p1);
 	t2 = *((const NODE *const *) p2);
@@ -2225,15 +2399,15 @@ comp_func(const void *p1, const void *p2)
 	t1 = force_string(t1);
 	t2 = force_string(t2);
 */
-	len1 = t1->stlen;
-	str1 = t1->stptr;
+	len1 = t1->ahname_len;
+	str1 = t1->ahname_str;
 
-	len2 = t2->stlen;
-	str2 = t2->stptr;
+	len2 = t2->ahname_len;
+	str2 = t2->ahname_str;
 
 	/* Array indexes are strings, compare as such, always! */
-	if (len1 == len2 || len1 < len2)
-		return memcmp(str1, str2, len1);
-	else
-		return memcmp(str1, str2, len2);
+	cmp1 = memcmp(str1, str2, len1 < len2 ? len1 : len2);
+	/* if prefixes are equal, size matters */
+	return (cmp1 != 0 ? cmp1 :
+		len1 < len2 ? -1 : (len1 > len2));
 }
