@@ -15,7 +15,7 @@
 #  endif
 #  define popen(c, m)	os_popen(c, m)
 #  define pclose(f)	os_pclose(f)
-   extern FILE *os_popen( char *, char * );
+   extern FILE *os_popen( char *, const char * );
    extern int  os_pclose( FILE * );
 # endif
 #endif
